import * as dotenv from 'dotenv';
dotenv.config();
import { loggerExpress, logger } from '@heathmont/node-logger';
import express, { NextFunction, Request, Response } from 'express';
import { timingSafeEqual } from 'crypto';
import { validationResult, body } from 'express-validator';
import { RestError, ErrorId } from './consts';
import axios from 'axios';
import bodyParser from 'body-parser';
import FormData from 'form-data';

const SERVICE_NAME = 'cloudflare-verifier';

logger.init({ prefix: SERVICE_NAME });

const app = express();

const areStringsEqual = (a: string, b: string) => {
  try {
    return timingSafeEqual(Buffer.from(a, 'utf8'), Buffer.from(b, 'utf8'));
  } catch {
    return false;
  }
};

app.use((req, res, next) => {
  const apiKey = req.headers['x-api-key'] || '';
  if (typeof apiKey === 'string' && apiKey.length > 0 && areStringsEqual(apiKey, process.env.API_KEY || '')) {
    next();
  } else {
    res.status(401).json({
      message: ErrorId.UNAUTHORIZED,
    });
  }
});

if (process.env.NODE_ENV === 'local') {
  app.use((_: Request, res: Response, next: NextFunction) => {
    res.header('Access-Control-Allow-Origin', '*');
    res.header('Access-Control-Allow-Credentials', 'true');
    res.header('Access-Control-Allow-Headers', 'Origin, X-Requested-With, Content-Type, Accept');
    res.header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS, PUT, DELETE');
    next();
  });
}

app.use(bodyParser.json());

app.use(loggerExpress);

const handlerWrapper =
  (controller: (req: Request, res: Response, next: NextFunction) => void) => async (req: Request, res: Response, next: NextFunction) => {
    try {
      const validationErrors = validationResult(req);
      if (!validationErrors.isEmpty()) {
        const invalidFields = validationErrors.array().map(({ param, value = '' }) => ({ param, value: value }));
        throw new RestError(400, ErrorId.INVALID_PARAMETERS, null, invalidFields);
      }
      return await controller(req, res, next);
    } catch (e) {
      next(e);
    }
  };

app.post(
  '/validate',
  body('token').isString(),
  handlerWrapper(async (req, res) => {
    const { token } = req.body;
    const requestBody = new FormData();
    requestBody.append('secret', process.env.CLOUDFLARE_SITE_SECRET);
    requestBody.append('response', token);
    const { data } = await axios.post(process.env.CLOUDFLARE_CHALLENGE_VERIFY_URL || '', requestBody);
    logger.info(`CF validation response ${JSON.stringify(data)}`);
    if (!data.success) {
      throw new RestError(400, ErrorId.FAILED_CLOUDFLARE_CHALLENGE, { cfResponse: data });
    }
    res.json({ token });
  }),
);

app.get('/healthcheck', (req, res) => {
  res.status(200).send('');
});

app.use((req: Request, res: Response, next: NextFunction) => {
  if (req.method === 'OPTIONS') {
    next();
  } else {
    logger.error(`Could not find resource for request: ${req.protocol}://${req.get('host')}${req.originalUrl}`);
    res.status(404).json({ id: ErrorId.RESOURCE_NOT_FOUND });
  }
});

// eslint-disable-next-line @typescript-eslint/no-unused-vars
app.use((err: Error, _: Request, res: Response, __: NextFunction) => {
  logger.error(err.message);
  if (err.stack) {
    logger.error(err.stack);
  }
  if (err instanceof RestError) {
    let errorResponse: { message: string; invalidFields?: { param: string; value: string }[] } = { message: err.message };
    if ('invalidFields' in err) {
      errorResponse = {
        ...errorResponse,
        invalidFields: err.invalidFields,
      };
    }
    if ('additionalData' in err) {
      errorResponse = {
        ...errorResponse,
        ...err.additionalData,
      };
    }
    res.status(err.status).json(errorResponse);
  } else {
    res.status(500).json({ id: ErrorId.INTERNAL_SERVER_ERROR });
  }
});

app.listen(process.env.SERVICE_PORT, () => {
  logger.info(`${SERVICE_NAME} listening on port ${process.env.SERVICE_PORT}`);
});
